/******************************************************************************
 * The copyright in this software is being made available under the BSD        *
 * License, included below. This software may be subject to other third party  *
 * and contributor rights, including patent rights, and no such rights are     *
 * granted under this license.                                                 *
 *    <OWNER> = ISO                                                            *
 *    <ORGANIZATION> = ISO                                                     *
 *    <YEAR> = 2015                                                            *
 *    Copyright (c) 2012-2015, ISO                                             *
 *                                                                             *
 * All rights reserved.                                                        *
 * Redistribution and use in source and binary forms, with or without          *
 * modification, are permitted provided that the following conditions are met: *
 *       * Redistributions of source code must retain the above copyright      *
 *         notice,this list of conditions and the following disclaimer.        *
 *       * Redistributions in binary form must reproduce the above             *
 *         copyright notice, this list of conditions and the following         *
 *         disclaimer in the documentation and/or other materials provided     *
 *         with the distribution.                                              *
 *       * Neither the name of the <ORGANIZATION> nor the names of its         *
 *         contributors may be used to endorse or promote products derived     *
 *         from this software without specific prior written permission.       *
 *                                                                             *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" *
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE   *
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE  *
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE   *
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR         *
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF        *
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS    *
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN     *
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)     *
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  *
 * POSSIBILITY OF SUCH DAMAGE.                                                 *
 ******************************************************************************/

#ifndef _JH_STREAM_H
#define _JH_STREAM_H

#include <stddef.h>
#include "jh_datatypes.h"
#include "jh_malloc.h"
#define JH_STREAM_BUFSIZE  (0xfffff)

/* memory stream implementation */
typedef struct jh_node_ts {
    struct jh_node_ts
                *next,
                *prev;
    unsigned char *buf;
    int owns_buf;   /* determine if the node owns the buffer */
    size_t
        buf_size,   /* size in bytes */
        data_size;  /* size of actual data in bytes */
} jh_node_t;

/* the destination manager is only to support IJG jpeg implementation
 * Other jpeg implementations may not require it */

typedef struct jh_mem_stream_ts {
    int (*read)(jh_stream_t* stream, unsigned char* buffer, size_t bytes);
    int (*write)(jh_stream_t* stream, unsigned char* buffer, size_t bytes);
    int (*seek)(jh_stream_t* stream, long long offset, unsigned origin);
    jh_buf_mgr_t* pri;    /* private interface to provide compatibility with IJG jpeg */
    size_t (*get_size)(jh_stream_t* strm);
    int (*empty)(jh_stream_t* stream);
    jh_node_t *head;
    jh_node_t *current;
    int offset;         /* byte offset in the current node to determine pointer of current reading/writing*/
} jh_mem_stream_t;

static jh_node_t* jh_create_node(const size_t sz, unsigned char* buf, jh_node_t *prev, jh_node_t *next) {
    jh_node_t
        *node = (jh_node_t*) jh_malloc(sizeof(jh_node_t));
    int own = 0;
    if (NULL == buf) { /* create buffer is not given */
        buf = (unsigned char*) jh_malloc(sz);
        own = 1;
    }
    if (NULL == node || NULL == buf) {
        jh_free(node);
        if (own) {
            jh_free(buf);
        }
        return NULL;
    }
    node->owns_buf = own;
    node->buf = buf;
    node->data_size = own ? 0 : sz;
    node->buf_size = sz;
    node->prev = prev;
    if (NULL != prev) {
        prev->next = node;
    }
    node->next = next;
    if (NULL != next) {
        next->prev = node;
    }
    return node;
}

static void jh_dispose_node(jh_node_t* n) {
    if(NULL != n) {
        if (n->owns_buf) {
            jh_free(n->buf);
        }
        jh_free(n);
        n = NULL;
    }
}

static int jh_on_init_impl(jh_stream_t *s_in, unsigned char **buf, size_t *buf_size) {
    jh_mem_stream_t *s = (jh_mem_stream_t*)s_in;
    if (NULL == s) {
        return 0; /* invalid argument */
    }
    if (NULL == s->head) { /* empty list - allocate */
        s->head = jh_create_node(JH_STREAM_BUFSIZE, NULL, NULL, NULL);
        if (NULL == s->head) {
            return 0; /* memory allocation failed */
        }
        s->current = s->head;
        s->offset = 0;
    }
    if (NULL != buf)
        *buf = s->current->buf + s->offset;
    if (NULL != buf_size)
        *buf_size = s->current->buf_size - s->offset;
    return 1;
}


static int jh_on_empty_imp(jh_stream_t *s_in, unsigned char **buf, size_t* buf_size) {
    jh_mem_stream_t *s = (jh_mem_stream_t*)s_in;
    if (NULL == s) {
        return 0;
    }
    s->current->data_size = s->current->buf_size;
    if (NULL == s->current->next) { /* allocate new node and data memory */
        size_t sz = 2*s->current->buf_size;
        sz = sz > JH_STREAM_BUFSIZE ? sz : JH_STREAM_BUFSIZE; /* set buffer size to maximum known */
        s->current->next = jh_create_node(sz, NULL, s->current, NULL);
        if (NULL == s->current->next) {
            return 0; /* memory allocation failed */
        }
    }
    s->current = s->current->next;
    s->offset = 0;
    if (NULL != buf)
        *buf = s->current->buf;
    if (NULL != buf_size)
        *buf_size = s->current->buf_size;
    return 1;
}

static int jh_on_term_impl(jh_stream_t* s_in, size_t free_in_buffer) {
    /* this can be empty if data_size maintained by the writer */
    jh_mem_stream_t *s = (jh_mem_stream_t*)s_in;
    if (NULL == s) {
        return 0; /* invalid arguments */
    }
    s->current->data_size = s->current->buf_size - free_in_buffer;
    return 1;
}


static int jh_mem_stream_write(jh_stream_t* s_in, unsigned char* buf, size_t bytes) {
    jh_mem_stream_t *s = (jh_mem_stream_t*)s_in;
    int bytes_left = (int)bytes; /* files are small so conversion is safe */
    int bytes_written = 0;
    if (NULL == s) { /* invalid argument */
        return 0;
    }
    if (NULL == s->head) {
        if (!s->pri->on_init(s_in, NULL, NULL)) {
            return 0;
        }
    }
    while (bytes_left) {
        int bytes_write;
        if ((int)s->current->buf_size <= s->offset) {
            if (!s->pri->on_empty(s_in, NULL, NULL)) {
                return 0;
            }
        }
        bytes_write = (int)s->current->buf_size - s->offset;
        bytes_write = bytes_write < bytes_left ? bytes_write : bytes_left;
        if (bytes_write < 0) {
            return bytes_written;
        }
        memcpy(s->current->buf + s->offset, (unsigned char*)buf, (size_t)bytes_write);
        s->offset += bytes_write;
        bytes_left -= bytes_write;
        buf += bytes_write;
        bytes_written += bytes_write;
        s->current->data_size += bytes_write;
    }
    return bytes_written;
}
/* pos is position of the byte (starts from 0) */
static int jh_mem_stream_seek(jh_stream_t* s_in, long long offset, unsigned origin) {
    jh_mem_stream_t *s = (jh_mem_stream_t*)s_in;
    int ret = 1;
    jh_node_t
            *p = NULL;
    if (NULL == s) {
        ret = 0;
        goto cleanup;
    }
    /* set starting point according to the requested origin */
    switch(origin) {
    case JH_SEEK_SET: /* reset to beginning */
        p = s->head;
        s->offset = 0;
        break;
    case JH_SEEK_CUR:
        p = s->current;
        break;
    case JH_SEEK_END:
        p = s->head;
        while(p && p->next) {
            p = p->next;
        }
        s->offset = (int)p->data_size-1;
        break; /* no more calculation required */
    default:
        ret = 0; /* error */
        goto cleanup;
    }
    /* find the node */
    if (offset >= 0) {
        while(p->next && (offset + s->offset) >  p->data_size-1) { /* should be p-> data_size in calculations */
            offset -= (p->data_size - s->offset);
            /* move to next block and set offset to its beginning */
            p = p->next;
            s->offset = 0;
        }
        /* set the stream to appropriate byte */
        s->current = p;
        if ((offset + s->offset) > p->data_size-1) { /* reach end of list */
            s->offset = (int)s->current->data_size; /* set to the end of the stream */
        } else {
            s->offset += (int)offset; /* remove long long since we know the files are small */
        }
    } else { /* negative offset */
        while(NULL != p->prev && -offset > s->offset) {
            offset += s->offset+1;
            s->offset = (int)p->prev->data_size - 1; /* set to the end of the buffer */
            p = p->prev;
        }
        /* set the stream to appropriate byte */
        if (-offset > s->offset) { /* reach head of list */
            s->current = s->head;
            s->offset = 0;
        } else {
            s->current = p;
            s->offset += (int)offset;
        }
    }
cleanup:
    return ret;
}

static int jh_mem_stream_read(jh_stream_t *s_in, unsigned char* buf, size_t bytes) {
    jh_mem_stream_t *s = (jh_mem_stream_t*)s_in;
    int bytes_left = (int)bytes;
    int bytes_read = 0;
    if (NULL == s || NULL == s->current) { /* invalid argument */
        return 0;
    }

    while (bytes_left) {
        int bytes_toread = 0;
        if ((int)s->current->data_size - 1 < s->offset) { /* we finish reading available buf */
            if (NULL != s->current->next) {
                s->current = s->current->next;
                s->offset = 0;
            } else {
                break;
            }
        }
        bytes_toread = (int)s->current->data_size - s->offset; /* available bytes to read */
        bytes_toread = bytes_toread < bytes_left ? bytes_toread : bytes_left;
        if (!bytes_toread) {
            break;
        }
        memcpy(buf, s->current->buf + s->offset, bytes_toread);
        s->offset += bytes_toread;
        bytes_left -= bytes_toread;
        buf += bytes_toread;
        bytes_read += bytes_toread;
    }
    return bytes_read;
}

static size_t jh_stream_size(jh_stream_t* strm) {
    jh_mem_stream_t* s = (jh_mem_stream_t*)strm;
    jh_node_t *p = NULL;
    size_t bytes = 0;
    if (NULL == strm)
        return 0;
    p = s->head;
    while(NULL != p) {
        bytes += p->data_size;
        p = p->next;
    }
    return bytes;
}

static int jh_empty_stream(jh_stream_t* strm) {
    jh_mem_stream_t* s = (jh_mem_stream_t*)strm;
    jh_node_t *p = NULL;
    if (NULL == strm) {
        return 0;
    }
    p = s->head;
    while (p) {
        jh_node_t *cur = p;
        p = p->next;
        jh_dispose_node(cur);
    }
    s->current = s->head = NULL;
    s->offset = 0;
    return 1;
}

/* helpers functions */
static void jh_init_stream(jh_mem_stream_t* strm, jh_buf_mgr_t* mgr) {
    /* stream initialization */
    mgr->on_empty = jh_on_empty_imp;
    mgr->on_init = jh_on_init_impl;
    mgr->on_term = jh_on_term_impl;
    strm->pri = mgr;
    strm->write = jh_mem_stream_write;
    strm->seek = jh_mem_stream_seek;
    strm->read = jh_mem_stream_read;
    strm->get_size = jh_stream_size;
    strm->empty = jh_empty_stream;
}

static void jh_attach_membuf(jh_mem_stream_t* strm, unsigned char* buf, size_t buf_size, size_t data_size) {
    jh_node_t zero_node = {0};
    strm->empty((jh_stream_t*)strm); /* may be skipped since it is empty in this implementation */
    strm->current = strm->head = (jh_node_t*)jh_malloc(sizeof(jh_node_t));
    *strm->head = zero_node;
    strm->head->buf = buf; /* take ownership */
    strm->head->buf_size = buf_size;
    strm->head->data_size = data_size;
    strm->offset = 0; /* it is already 0 but for clarity sake */
}

static void jh_detach_membuf(jh_mem_stream_t* strm) {
    if (NULL != strm && NULL != strm->head) {
        strm->head->buf = NULL; /* we know that in this implementation only one buffer exists */
    }
}

#endif /*_JH_STREAM_H*/
