/******************************************************************************
 * The copyright in this software is being made available under the BSD        *
 * License, included below. This software may be subject to other third party  *
 * and contributor rights, including patent rights, and no such rights are     *
 * granted under this license.                                                 *
 *    <OWNER> = ISO                                                            *
 *    <ORGANIZATION> = ISO                                                     *
 *    <YEAR> = 2015                                                            *
 *    Copyright (c) 2012-2015, ISO                                             *
 *                                                                             *
 * All rights reserved.                                                        *
 * Redistribution and use in source and binary forms, with or without          *
 * modification, are permitted provided that the following conditions are met: *
 *       * Redistributions of source code must retain the above copyright      *
 *         notice,this list of conditions and the following disclaimer.        *
 *       * Redistributions in binary form must reproduce the above             *
 *         copyright notice, this list of conditions and the following         *
 *         disclaimer in the documentation and/or other materials provided     *
 *         with the distribution.                                              *
 *       * Neither the name of the <ORGANIZATION> nor the names of its         *
 *         contributors may be used to endorse or promote products derived     *
 *         from this software without specific prior written permission.       *
 *                                                                             *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" *
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE   *
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE  *
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE   *
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR         *
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF        *
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS    *
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN     *
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)     *
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  *
 * POSSIBILITY OF SUCH DAMAGE.                                                 *
 ******************************************************************************/

#include <string.h>
#include <stdio.h>
#include <math.h>
#include "jh_helpers.h"
#include "jh_color.h"
#include "jh_datatypes.h"
#include "jh_defs.h"


const float default_csb[JH_CS_MATRIX_SIZE] =
{
    1.0f, 0.0f, 0.0f,
    0.0f, 1.0f, 0.0f,
    0.0f, 0.0f, 1.0f
};

const float default_csr[JH_CS_MATRIX_SIZE] =
{
    1.0f,  0.0f,      1.402f,
    1.0f, -0.34414f, -0.71414f,
    1.0f,  1.772f,    0.0f
};

void jh_get_default_bga(float bga[JH_CS_TRC_SIZE])
{
    int i;
    for(i=0; i<JH_CS_TRC_SIZE; i++) {
        bga[i] = default_bga[i];
    }
    return;
}

static int jh_is_invalid_colorspace_matrix(const float *cs_mat_3x3, jh_error_t *err) {
    int i=0;
    if (cs_mat_3x3 == NULL) {
        jh_set_error(err, JH_INVALID_ARGUMENT, "NULL colorspace matrix");
        return 1;
    }
    for(i=0; i<JH_CS_MATRIX_SIZE; i++) {
        if ((cs_mat_3x3[i] < -30.0f) || (cs_mat_3x3[i] > 30.0f)) {
            jh_set_error(err, JH_INVALID_ARGUMENT, "colorspace matrix value %f is out of range [-30.0, 30.0]", cs_mat_3x3[i]);
            return 1;
        }
    }
    return 0;
}

static int jh_is_invalid_colorconvert_params(const jh_colorconvert_params_t * p, jh_error_t *err) {
    if (p == NULL) {
        jh_set_error(err, JH_INVALID_ARGUMENT, "NULL colorspace parameters");
        return 1;
    }
    if(jh_is_invalid_colorspace_matrix(p->conv_matrix, err)) {
        jh_set_error(err, JH_INVALID_ARGUMENT, "Invalid colorspace conversion matrix");
        return 1;
    }
    return 0;
}


static int jh_is_zero_array(const float *array, const unsigned int length)
{
    unsigned int i;
    for (i=0U; i<length; i++) {
        if(fabs(array[i]) > JH_MIN_FLOAT) {
            return 0;
        }
    }
    return 1;
}

float jh_hdr_lum(const float R, const float G, const float B, const float *coefs)
{
    float lum = 0.0f;
    if(NULL != coefs) {
        lum = coefs[0] * R + coefs[1] * G + coefs[2] * B;
    }
    else {
        lum = 0.299f * R + 0.587f * G + 0.114f * B;
    }
    return ((lum >= JH_MIN_LUM) ? lum : JH_MIN_LUM);
}

int jh_init_colorconvert_params(jh_colorconvert_params_t *csp,
                                const float *cs_conv_mat,
                                const float *base_trc)
{
    int i;
    if(jh_is_zero_array(cs_conv_mat, JH_CS_MATRIX_SIZE)) {
        memcpy(&(csp->conv_matrix[0]), default_csb, sizeof(float) * JH_CS_MATRIX_SIZE);
    }
    else {
        memcpy(&(csp->conv_matrix[0]), cs_conv_mat, sizeof(float) * JH_CS_MATRIX_SIZE);
    }
    if(jh_is_zero_array(base_trc, JH_CS_TRC_SIZE)) {
        jh_get_default_bga(csp->base_trc);
    }
    else {
        memcpy(&(csp->base_trc[0]), base_trc, sizeof(float) * JH_CS_TRC_SIZE);
    }
    /* enforce min value in TRC */
    if (csp->base_trc[0] < 1.e-7) {
        for(i=0; i<JH_CS_TRC_SIZE; i++) {
            csp->base_trc[i] = csp->base_trc[i] + 1.f/(1<<JH_BITS);
        }
    }
    return 0;
}

/* convert an 8-bit sRGB pixel to destination linear RGB float value */
void jh_baseRGB_to_hdrRGB(const unsigned char *s, float *d, const jh_colorconvert_params_t* csp)
{
    const float *mat = &(csp->conv_matrix[0]);
    const float *trc = &(csp->base_trc[0]);
    float r, g, b;
    r = trc[s[0]];
    g = trc[s[1]];
    b = trc[s[2]];
    d[0] = mat[0] * r + mat[1] * g + mat[2] * b;
    d[1] = mat[3] * r + mat[4] * g + mat[5] * b;
    d[2] = mat[6] * r + mat[7] * g + mat[8] * b;
}
