/******************************************************************************
 * The copyright in this software is being made available under the BSD        *
 * License, included below. This software may be subject to other third party  *
 * and contributor rights, including patent rights, and no such rights are     *
 * granted under this license.                                                 *
 *    <OWNER> = ISO                                                            *
 *    <ORGANIZATION> = ISO                                                     *
 *    <YEAR> = 2015                                                            *
 *    Copyright (c) 2012-2015, ISO                                             *
 *                                                                             *
 * All rights reserved.                                                        *
 * Redistribution and use in source and binary forms, with or without          *
 * modification, are permitted provided that the following conditions are met: *
 *       * Redistributions of source code must retain the above copyright      *
 *         notice,this list of conditions and the following disclaimer.        *
 *       * Redistributions in binary form must reproduce the above             *
 *         copyright notice, this list of conditions and the following         *
 *         disclaimer in the documentation and/or other materials provided     *
 *         with the distribution.                                              *
 *       * Neither the name of the <ORGANIZATION> nor the names of its         *
 *         contributors may be used to endorse or promote products derived     *
 *         from this software without specific prior written permission.       *
 *                                                                             *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" *
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE   *
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE  *
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE   *
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR         *
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF        *
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS    *
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN     *
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)     *
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  *
 * POSSIBILITY OF SUCH DAMAGE.                                                 *
 ******************************************************************************/

#include "jinclude.h"
#include "jerror.h"
#include "jh_stream_datadst.h"

/* implementation of destination manager in terms of jpeg-hdr stream */
/* Comments and function description are from IJG jpeg library sample */


static void jh_init_strm_destination (j_compress_ptr cinfo);
static boolean jh_empty_strm_output_buffer (j_compress_ptr cinfo);
static void jh_term_strm_destination (j_compress_ptr cinfo);

/* implementation of destination manager in terms of jpeg-hdr stream */
/* Comments and function description are from IJG jpeg library sample */

typedef struct {
    struct jpeg_destination_mgr pub;  /* public fields */
    jh_stream_t *strm;                /* stream interface */
} jh_stream_destination_mgr;


static void jh_init_strm_destination (j_compress_ptr cinfo)
{
    jh_stream_destination_mgr* dest = (jh_stream_destination_mgr*) cinfo->dest;
    jh_stream_t *strm = dest->strm;
    if (!strm->pri->on_init(strm, &dest->pub.next_output_byte, &dest->pub.free_in_buffer))
        ERREXIT1(cinfo, JERR_OUT_OF_MEMORY, 10);
}


/* Based on IJG sample files
 * jh_empty_strm_output_buffer --- called whenever buffer fills up.
 *
 * In typical applications, this should write the entire output buffer
 * (ignoring the current state of next_output_byte & free_in_buffer),
 * reset the pointer & count to the start of the buffer, and return TRUE
 * indicating that the buffer has been dumped.
 *
 * In applications that need to be able to suspend compression due to output
 * overrun, a FALSE return indicates that the buffer cannot be emptied now.
 * In this situation, the compressor will return to its caller (possibly with
 * an indication that it has not accepted all the supplied scanlines).  The
 * application should resume compression after it has made more room in the
 * output buffer.  Note that there are substantial restrictions on the use of
 * suspension --- see the documentation.
 *
 * When suspending, the compressor will back up to a convenient restart point
 * (typically the start of the current MCU). next_output_byte & free_in_buffer
 * indicate where the restart point will be if the current call returns FALSE.
 * Data beyond this point will be regenerated after resumption, so do not
 * write it out when emptying the buffer externally.
 */

static boolean jh_empty_strm_output_buffer (j_compress_ptr cinfo)
{
  jh_stream_destination_mgr* dest = (jh_stream_destination_mgr*) cinfo->dest;
  jh_stream_t *strm = dest->strm;
  if (!strm->pri->on_empty(strm, &dest->pub.next_output_byte, &dest->pub.free_in_buffer))
      ERREXIT1(cinfo, JERR_OUT_OF_MEMORY, 10);
  return TRUE;
}


/*
 * Terminate destination --- called by jpeg_finish_compress
 * after all data has been written.  Usually needs to flush buffer.
 *
 * NB: *not* called by jpeg_abort or jpeg_destroy; surrounding
 * application must deal with any cleanup that should happen even
 * for error exit.
 */


static void jh_term_strm_destination (j_compress_ptr cinfo)
{
    jh_stream_destination_mgr* dest = (jh_stream_destination_mgr*) cinfo->dest;
    jh_stream_t *strm = dest->strm;
    if (!strm->pri->on_term(strm, dest->pub.free_in_buffer)) {
        ERREXIT1(cinfo, JERR_FILE_READ, 0); /* NOTE: the error reported is not correct but the only available in jpeg */
    }

}


/*
 * Prepare for output to a memory buffer.
 * The caller may supply an own initial buffer with appropriate size.
 * Otherwise, or when the actual data output exceeds the given size,
 * the library adapts the buffer size as necessary.
 * The standard library functions malloc/free are used for allocating
 * larger memory, so the buffer is available to the application after
 * finishing compression, and then the application is responsible for
 * freeing the requested memory.
 */
void jh_stream_dest(j_compress_ptr cinfo, jh_stream_t *strm)
{
  jh_stream_destination_mgr
              *dest,
              dest_zero = {{0}};

  if (strm == NULL) /* sanity check */
    ERREXIT(cinfo, JERR_BUFFER_SIZE); /* NOTE: we use this error code to indicate NULL stream */

  /* The destination object is made permanent so that multiple JPEG images
   * can be written to the same buffer without re-executing jpeg_mem_dest.
   */
  if (cinfo->dest == NULL) {    /* first time for this JPEG object? */
    cinfo->dest = (struct jpeg_destination_mgr *)
      (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
                  SIZEOF(jh_stream_destination_mgr));
  }

  dest = (jh_stream_destination_mgr*) cinfo->dest;
  *dest = dest_zero;
  dest->pub.init_destination = jh_init_strm_destination;
  dest->pub.empty_output_buffer = jh_empty_strm_output_buffer;
  dest->pub.term_destination = jh_term_strm_destination;
  dest->strm = strm;
}

